import cv2
import numpy as np
import glob
from PIL import Image
from typing import List, Tuple, Union


def xy_to_homo(x:Union[float, int], 
               y:Union[float, int], 
               homo:np.array) -> Tuple:
    # homo in shape of [3,3]
    
    img_coord = np.array([x, y, 1])
    homo_coord = np.matmul(homo, img_coord)
    x_homo, y_homo, _ = homo_coord/homo_coord[2]
    
    return round(x_homo), round(y_homo)


def homo_to_xy(x_homo:Union[float, int],
               y_homo:Union[float, int], 
               homo:np.array) -> Tuple:
    # homo in shape of [3,3]
    
    homo_coord = np.array([x_homo, y_homo, 1])
    xy_coord = np.matmul(np.linalg.inv(homo), homo_coord)
    x, y, _ = xy_coord/xy_coord[2]
    
    return round(x), round(y)



import numpy as np

def box_iou(box1, box2):
    # https://github.com/pytorch/vision/blob/master/torchvision/ops/boxes.py
    """
    Return intersection-over-union (Jaccard index) of boxes.
    Both sets of boxes are expected to be in (x1, y1, x2, y2) format.
    Arguments:
        box1 (Tensor[N, 4])
        box2 (Tensor[M, 4])
    Returns:
        iou (Tensor[N, M]): the NxM matrix containing the pairwise
            IoU values for every element in boxes1 and boxes2
    """

    def box_area(box):
        # box = 4xn
        return (box[2] - box[0]) * (box[3] - box[1])

    area1 = box_area(box1.T)
    area2 = box_area(box2.T)

    # inter(N,M) = (rb(N,M,2) - lt(N,M,2)).clamp(0).prod(2)
    inter = (np.minimum(box1[:, 2:], box2[:, 2:]) - np.maximum(box1[:, :2], box2[:, :2])).clip(0).prod(1)
    return inter / (area1[:, None] + area2 - inter)  # iou = inter / (area1 + area2 - inter)



import os
from typing import List, Tuple
from ODToolkit._utils import box_iou
import numpy as np


def get_fpfn(gt_lbl: List, 
             pred_lbl: List, 
             img_w,
             img_h,
             iou_thresh: float) -> Tuple:
        FP = []
        FN = []
        correct = []
        pred_root = os.path.split(pred_lbl[0])[0]

        for i, _ in enumerate(gt_lbl):
            with open(gt_lbl[i], 'r') as f:
                gt_lines = f.readlines()
                gt_box = []
                for line in gt_lines:
                    cls, xc, yc, w, h = line.split(' ')

                    xc, yc, h, w = float(xc), float(yc), float(h), float(w)
                    xc, yc, h, w = int(img_w*xc), int(img_h*yc), int(img_h*h), int(img_w*w)

                    xc, yc, x2, y2 = xc-w//2, yc-h//2, xc+w//2, yc+h//2
                    gt_box.append([xc, yc, x2, y2])

            path, fname = os.path.split(gt_lbl[i])
            pred_file = pred_root+'/'+fname

            pred_box = []
            if pred_file in pred_lbl:
                with open(pred_file, 'r') as f:
                    lines = f.readlines()
                    for line in lines:
                        cls, conf, xc, yc, w, h = line.split(' ')
                        xc, yc, h, w = float(xc), float(yc), float(h), float(w)
                        xc, yc, h, w = int(img_w*xc), int(img_h*yc), int(img_h*h), int(img_w*w)

                        xc, yc, x2, y2 = xc-w//2, yc-h//2, xc+w//2, yc+h//2
                        pred_box.append([xc, yc, x2, y2])

            # compare iou and find out fp and fn
            if pred_box == [] and gt_box != []:
                for gtb in gt_box:
                    FN.append(gtb)
                    xc, yc, x2, y2 = gtb

            elif gt_box == [] and pred_box != []:
                for pdb in pred_box:
                    FP.append(pdb)
            else:      
                for pdb in pred_box:
                    iou_fp = []
                    for gtb in gt_box:
                        iou_val = box_iou(np.array(pdb).reshape(1,4), np.array(gtb).reshape(1,4))
                        iou_fp.append(iou_val.item())
                    iou_fp = np.array(iou_fp)
                    if all(i < iou_thresh for i in iou_fp): # false positive
                        FP.append(pdb)
                    else:
                        # means there is a gt box satisfy the iou_thresh
                        correct.append(pdb)

                for gtb in gt_box:
                    iou_fn = []
                    for pdb in pred_box:
                        iou_val = box_iou(np.array(pdb).reshape(1,4), np.array(gtb).reshape(1,4))
                        iou_fn.append(iou_val.item())
                    iou_fn = np.array(iou_fn)
                    if all(i < iou_thresh for i in iou_fn): # false negative
                        xc, yc, x2, y2 = gtb
                        FN.append(gtb)
                        
        return FP, FN, correct


import glob
import numpy as np
from PIL import Image
import os
import cv2
from typing import List, Tuple


def valid_x(x_target, x1, x2):
    '''
    x1-x2 represents width or height
    thus, for a valid x proposal, we just want to make sure there is no overlapping
    '''
    
    diff = x2-x1
    if (x_target+diff) < x1 or x_target > x2:
        return True
    else:
        return False
    
    
def copy_paste_augmentation(train_img: List[str], 
                            train_lbl: List[str], 
                            K:int, 
                            savepath_img:str, 
                            savepath_lbl:str, 
                            R: int=1,
                            width:int,
                            height:int,
                            area_threshold:float,
                            mask_percentage:float,
                            mask_top=True,
                            quit: int=100):

    THRESHOLD_AREA = int(area_threshold*width*height)
    margin = int(mask_percentage*height)
    area = width*height

    with open(train_lbl[K], 'r') as f:
        lines = f.readlines()

    # some quick return algorithm
    # 1. if no label, then directly return
    # 2. if all small boxes, directly return
    if lines == []:
        return
    big_box = 0
    for line in lines:
        cls, xc, yc, w, h = line.split(' ')
        xc, yc, h, w = float(xc), float(yc), float(h), float(w)
        xc, yc, h, w = int(width*xc), int(height*yc), int(height*h), int(width*w)
        
        if h*w <= THRESHOLD_AREA:
            continue
        else:
            big_box+=1
    if big_box == 0:
        return
    
    # if there are some big boxes in the image, process it and create aug files
    image = np.array(Image.open(train_img[K]))
    new_im = image.copy()

    occ = [] # put all boxes in to this list
    with open(savepath_lbl, 'w') as f:
        f.write('')
        
    for line in lines:
        with open(savepath_lbl, 'a') as f:
            f.write(line)
            
        cls, xc, yc, w, h = line.split(' ')

        xc, yc, h, w = float(xc), float(yc), float(h), float(w)
        xc, yc, h, w = int(width*xc), int(height*yc), int(height*h), int(width*w)

        y1 = max(0, yc - h/2)
        x1 = max(0, xc - w/2)
        y2 = min(height, y1+h)
        x2 = min(width, x1+w)
        
        if h*w <= THRESHOLD_AREA:
            continue
        else:
            occ.append([cls, x1, y1, x2, y2]) # x1y1x2y2
            
    # now we make sure in occ and occ_new, all are big boxes
    occ_new = occ.copy()
    if len(occ) > 3: # if too many boxes, we don't copy-paste that much
        R = 1
    for _ in range(R):
        for box in occ:
            _, x1, y1, x2, y2 = box
            h, w = y2-y1, x2-x1
            
            valid = False
            val_cnt = 0 # check if while loop is a endless loop
            while not valid: # keep finding valid position to place the fake box
                dx = np.random.randint(0, width-w)
                if mask_top:
                    dy = np.random.randint(margin, height-h)
                else:
                    dy = np.random.randint(0, margin-h)
                    
                res = []
                val_cnt+=1
                if val_cnt > quit: # if bad-luck, we give up this box
                    break 
                    
                for boxn in occ_new:
                    cls, x1n, y1n, x2n, y2n = boxn
                    hn, wn = y2n-y1n, x2n-x1n
                    
                    res.append(valid_x(dx, x1n, x2n) or valid_x(dy, y1n, y2n))

                if np.all(np.array(res, dtype=bool)):
                    valid = True
                    with open(savepath_lbl, 'a') as f:
                        # cls, xc, yc, w, h
                        line = ' '.join([str(i) for i in [cls, (dx+w/2)/width, (dy+h/2)/height, w/width, h/height]])
                        f.write(line+'\n')
                

            if val_cnt <=quit: # if lucky, we move to new round, if not, we give up this box
                patch = image[int(y1):int(y1+h), int(x1):int(x1+w), :]
                
                if np.random.rand() < 0.5: # vertical random flip
                    patch = patch[::-1, :, :]
                if np.random.rand() < 0.5: # horizontal random flip
                    patch = patch[:, ::-1, :]
                    
                new_im[int(dy):int(dy+h), int(dx):int(dx+w), :] = patch

                # update occ_new in x1y1x2y2
                occ_new.append([cls, dx, dy, dx+w, dy+h])
    
    image_ready = Image.fromarray(new_im)
    image_ready.save(savepath_img)
    
    return image_ready